""" Constants file for ThermoFisher Qubit 4 Adapter"""

# Instrument Software Details
QUBIT_SOFTWARE = "Qubit 4 software"
MODEL_NUMBER = "Qubit 4"
BRAND_NAME = "Qubit"
PRODUCT_MANUFACTURER = "Thermo Fisher Scientific"
DEVICE_TYPE = "fluorescence detector"
DISPLAY_NAME = "Thermo Fisher Qubit 4"

# Errors
VALUE_ERROR = "Unable to find value for column"
UNSUPPORTED_FILE_FORMAT_ERROR = "Unsupported file format. Expected xlsx. Actual: "
