NUCLEOCOUNTER_SOFTWARE_NAME = "NucleoView"
NUCLEOCOUNTER_DEVICE_TYPE = "dark field imager (cell counter)"
NUCLEOCOUNTER_DETECTION_TYPE = "dark field"

DEFAULT_ANALYST = "NC-200"
DEFAULT_MODEL_NUMBER = "NucleoCounter NC-200"
DEFAULT_EPOCH_TIMESTAMP = "1970-01-01"
