DEVICE_TYPE = "plate reader"
# The epoch is a place holder for measurement time since is not optional but
# the input file does not include a time that can be used.
EPOCH = "1970-01-01T00:00:00-00:00"
REDUCED = "Reduced"
