from allotropy.__about__ import __version__

ASM_CONVERTER_NAME = "allotropy"
ASM_CONVERTER_VERSION = __version__

# Value to use for encoding str params when we want to use chardet to detect the encoding.
CHARDET_ENCODING = "chardet"

DEFAULT_ENCODING = "UTF-8"
