ASM_CONVERTER_NAME = "allotropy_biorad_bioplex_manager"
WELLS_TAG = "Wells"
ANALYTE_NAME = "AnalyteName"
MW_ANALYTES = "MWAnalytes"
RP1_GAIN = "RP1Gain"
LABEL = "Label"
RUN_CONDITIONS = "RunConditions"
MACHINE_INFO = "MachineInfo"
DOC_LOCATION_TAG = "NativeDocumentLocation"
DESCRIPTION_TAG = "Description"
PLATE_DIMENSIONS_TAG = "PlateDimensions"
TOTAL_WELLS_ATTRIB = "TotalWells"
VERSION_ATTRIB = "BioPlexManagerVersion"
REGION_COUNT = "RegionCount"
SAMPLES = "Samples"
BEAD_MAP = "BeadMap"
MEDIAN = "Median"
MEMBER_WELLS = "MemberWells"
ROW_NUMBER = "RowNo"
COLUMN_NUMBER = "ColNo"
SERIAL_NUMBER = "SerialNumber"
WELL_NUMBER = "WellNumber"
WELL_NO = "WellNo"
REGION_NUMBER = "RegionNumber"
REGIONS_OF_INTEREST = "RegionsOfInterest"
STOP_READING_CRITERIA = "StopReadingCriteria"
READING = "Reading"
RUN_SETTINGS = "RunSettings"
SAMPLE_VOLUME = "SampleVolume"
CODE = "Code"
DILUTION = "Dilution"
BEAD_COUNT = "BeadCount"
TOTAL_EVENTS = "TotalEvents"
ACQ_TIME = "AcquisitionTime"
BEAD_REGIONS = "BeadRegions"
PRODUCT_MANUFACTURER = "Bio-Rad"
SOFTWARE_NAME = "“Bio-Plex Manager"
DEVICE_TYPE = "multi analyte profiling analyzer"
USER = "User"
CONTAINER_TYPE = "well plate"

ROW_NAMES = "ABCDEFGH"

ERROR_MAPPING = {
    "1": "Low bead number",
    "2": "Aggregated beads",
    "3": "Classify efficiency",
    "4": "Region selection",
    "5": "Platform temperature",
}
